/*
  lp_GT832E_01_BME
  (c)2022 Forward Computing and Control Pty. Ltd.
   This code is not warranted to be fit for any purpose. You may only use it at your own risk.
   This code may be freely used for both private and commercial use
   Provide this copyright is maintained.
*/

// advertising ~39uA  for GT832E_01 + BME280
// sleeping ~2.4uA  for GT832E_01 + BME280
//  Battery life for 10sec advert every 100sec
// GT832E_01 Crystal LowFreq clock:  10sec advert in 100sec => 39*10/100 + 2.4*(100-10)/100 = ~6.06uA  for uC using Crystal
// Using a CR2032 235mAh to 2V from https://data.energizer.com/pdfs/cr2032_eu.pdf
// i.e. CR2032 gives 235e-3/6.06e-6 hrs = 38779hrs => 1616 days => ~4 years 5months for GT832E_01 module

const char LOCAL_NAME[] = "W_1"; //
//  say T_.. for Temp(C) sensors, H_.. for Temp(C),Humditiy(%) sensors, W_.. for Temp(C),Humdity(%),Barometer(hPa) sensors, WL_.. for Temp(C),Humdity(%),Barometer(hPa),Lux level(Lux)
//  keep this LOCAL_NAME short as it uses up available advertising data space (see below)
// Format of advertized name
// For T_.. devices   LOCAL_NAME,degC
// For H_.. devices   LOCAL_NAME,degC,RH%
// For W_.. devices   LOCAL_NAME,degC,RH%,hPa
// For WL_.. devices  LOCAL_NAME,degC,RH%,hPa,Lux

#include "SafeString.h"
#include "BLEPeripheral.h"
#include <Wire.h>

#include "SparkFunBME280.h"
BME280 BME_Sensor;


float luxLevel = 0.0;
float tempC = 1000.0; // invalid
float hPa = 0.0;
float RH = 0.0;

BLEPeripheral ble;

// BLE_SCAN_DATA_MAX_VALUE_LENGTH == 29 in BLEDeviceLimits.h
// to this is added length and type for total of 31 == max advertisize packet size
// here we also set flags for not connectable which uses 3 bytes length(1), type(1), flags(1)
// so MAX_SCAN_DATA_LEN reduced to 26
const size_t MAX_DATA_LEN = 26;
// this can be split between say device NAME and Manufacturing Data
// but that uses another two bytes for the length and type bytes for the Manufacturing Data
//
// For simplicity and readability here all the data is added to the NAME as ascii date comma separated
// If using NAME and  Manufacturing Data then sending data as binary in  Manufacturing Data allows for more data.
//
// Note could also just have  Manfacturing Data with out a device NAME but easier to read data in the name with scanning apps

// uncomment this to enable Serial output uses a lot of supply current
//#define DEBUG
bool BME_SensorInitialized = false;
cSF(sfData, MAX_DATA_LEN); // Local name and formatted measurements SafeString will limit data to this length. Excess data is dropped NOT truncated see SafeString docs/tutorials

unsigned short ADVERT_INTERVAL = 750; // default it 500, 750ms gives 13 advert packets in the 10 secs

lp_timer BLE_Timer; // the advert timer
// 10sec advert on 90sec advert off
const unsigned long BLE_ADVERT_MS = 10ul * 1000; // advertise for 10sec at 750ms intervals i.e. ~13 advert packets
const unsigned long BLE_NO_ADVERT_MS = 90ul * 1000; // go silent for 90sec

// 10sec advert on 20sec advert off
//const unsigned long BLE_ADVERT_MS = 10ul * 1000; // advertise for 10sec at 750ms intervals i.e. ~13 advert packets
//const unsigned long BLE_NO_ADVERT_MS = 20ul * 1000; // go silent for 20sec

cSF(sfErrorStr, 80);

lp_timer BME_ReadResultDelay;
const unsigned long BME_READ_RESULT_DELAY_TIME = 25; // ms for temp,rh,hPa

// this method set the advertized NAME to include the data comma delimited
void setAdvertData() {
  sfData = LOCAL_NAME;
  sfData += ',';
  sfData.print(tempC, 1);
  sfData += ',';
  sfData.print(RH, 0);
  sfData += ',';
  sfData.print(hPa, 1);
  ble.setAdvertisedName(sfData.c_str());  //
#ifdef DEBUG
  Serial.print("set data to ");
  Serial.println(sfData);
#endif

}

void setup() {
#ifdef DEBUG
  Serial.setPins(30, 29); // Rx 30 == D0, Tx 29 == D1
  Serial.begin(115200);
  Serial.println();
  for (int i = 10; i > 0; i--) {
    Serial.print(' '); Serial.print(i);
    delay(500);
  }
  Serial.println();
#endif

  Wire.setPins(8, 6); // sda 8, clk 6
  Wire.begin();

  // Begin the BME280 using the Wire I2C port
  // .begin will return true on success, or false on failure to communicate
  if (BME_Sensor.beginI2C()) {
#ifdef DEBUG
    Serial.println("Found BME_Sensor");
#endif
    BME_SensorInitialized = true; // else leave as false
  } else {
    sfErrorStr += "Error: BME_Sensor not found\n";
  }
  BME_Sensor.setMode(MODE_SLEEP); //Sleep for now

#ifdef DEBUG
  Serial.println(sfErrorStr);
#endif

  // set advertised name
  ble.setConnectable(false);
  ble.setTxPower(+4);
  ble.setAdvertisingInterval(ADVERT_INTERVAL);
  setAdvertData();  // set initial null values and name
  startBME_Conversion();

#ifdef DEBUG
  Serial.println(" BLE started");
#endif

#ifdef DEBUG
  Serial.println("setup() finished");
#endif
}

void stopAdvert() {
  ble.end();
#ifdef DEBUG
  Serial.print("stop ble "); Serial.println(millis());
#endif
  BLE_Timer.startDelay(BLE_NO_ADVERT_MS, startBME_Conversion); //wait 90sec before next reading and restarting advert
}

void restartAdvert() {
  setAdvertData();  // pick up latest values
  ble.begin();
#ifdef DEBUG
  Serial.print("start ble ");; Serial.println(millis());
#endif
  BLE_Timer.startDelay(BLE_ADVERT_MS, stopAdvert); // stop in 10sec
}


bool BME_ConversionRunning = false;
void startBME_Conversion() {
  if (!BME_ConversionRunning) {
    BME_ConversionRunning = true;
    if (BME_SensorInitialized) {
      BME_Sensor.setMode(MODE_FORCED); //Wake up sensor and take reading
    }
  } // else just schedule to pickup last result
  BME_ReadResultDelay.startDelay(BME_READ_RESULT_DELAY_TIME, readBME_Results);
}

void readBME_Results() {
  if (BME_SensorInitialized) {
    BME_ConversionRunning = false;
    tempC = BME_Sensor.readTempC();
    RH = BME_Sensor.readFloatHumidity();
    hPa = BME_Sensor.readFloatPressure() / 100;
  } else {
    tempC = 1000;
    RH = 0;
    hPa = 0;
  }
#ifdef DEBUG
  Serial.print(" Temp: ");
  Serial.print(tempC, 1);
  Serial.print(" Humidity: ");
  Serial.print(RH, 0);
  Serial.print(" Pressure: ");
  Serial.print(hPa, 1);
#endif
  restartAdvert();
}

void loop() {
  sleep(); // just sleep here waiting for the timer to trigger
}
